from ast import literal_eval

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import Parameter
from bw2data.parameters import ProjectParameter, DatabaseParameter, ActivityParameter, parameters
from activity_browser.ui.icons import qicons

from .parameter_new_automatic import ParameterNewAutomatic


class ParameterNewFromParameter(ABAction):
    """
    ABAction to create a new Parameter from an instantiated Parameter namedtuple
    """

    icon = qicons.add
    text = "New parameter..."

    @staticmethod
    @exception_dialogs
    def run(parameter: Parameter):
        if not isinstance(parameter, Parameter) or parameter.param_type is None:
            raise ValueError("Parameter must be an instance of Parameter")

        if not parameter.name.isidentifier():
            raise ValueError("Parameter name must be a valid Python identifier")

        # select the right group and instruct the controller to create the parameter there
        if parameter.param_type == "project":
            ProjectParameter(
                name=parameter.name,
                formula=parameter.data.get("formula", None),
                amount=parameter.amount,
                data=parameter.data,
            ).save()
        elif parameter.param_type == "database":
            DatabaseParameter(
                database=parameter.group,
                name=parameter.name,
                formula=parameter.data.get("formula", None),
                amount=parameter.amount,
                data=parameter.data,
            ).save()
        elif parameter.param_type == "activity":
            mock = ActivityParameter.get_or_none(group=parameter.group)
            if mock is None:
                ParameterNewAutomatic.run([int(parameter.group)])
                mock = ActivityParameter.get(group=parameter.group)

            ActivityParameter(
                group=parameter.group,
                database=mock.database,
                code=mock.code,
                name=parameter.name,
                formula=parameter.data.get("formula", None),
                amount=parameter.amount,
                data=parameter.data,
            ).save()

        parameters.recalculate()




from qtpy import QtWidgets
from qtpy.QtCore import Qt

import bw2data as bd
import pandas as pd

from activity_browser import actions
from activity_browser.ui import widgets, icons, delegates
from activity_browser.bwutils import AB_metadata, is_node_product


class FunctionalUnitSection(QtWidgets.QWidget):
    def __init__(self, calculation_setup_name: str, parent=None):
        super().__init__(parent)

        self.calculation_setup_name = calculation_setup_name
        self.calculation_setup = bd.calculation_setups.get(self.calculation_setup_name)

        self.view = FunctionalUnitView()
        self.model = FunctionalUnitModel()
        self.view.setModel(self.model)

        self.build_layout()

    def build_layout(self):
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        self.setLayout(layout)

    def sync(self):
        try:
            self.calculation_setup = bd.calculation_setups[self.calculation_setup_name]
            self.model.setDataFrame(self.build_df())
        except KeyError:
            self.parent().close()
            self.parent().deleteLater()

    def build_df(self):
        keys, amounts = [], []
        cols = ["unit", "name", "product", "location", "database", "processor"]

        for fu in self.calculation_setup.get("inv", []):
            for key, amount in fu.items():
                keys.append(key)
                amounts.append(amount)

        act_df = AB_metadata.get_metadata(keys, cols)
        act_df["amount"] = amounts
        act_df["_activity_key"] = keys
        act_df["_cs_name"] = self.calculation_setup_name

        act_df["_processor_key"] = act_df["processor"]
        act_df["_processor_key"] = act_df["_processor_key"].fillna(act_df["_activity_key"])

        # Retrieve metadata for unique processor keys, focusing on the "name" column.
        processor_df = AB_metadata.get_metadata(act_df["_processor_key"].unique(), ["name"])

        # Flatten the index of the processor DataFrame to ensure compatibility with merging.
        processor_df.index = processor_df.index.to_flat_index()

        # Merge the processor keys from the activity DataFrame with the processor metadata.
        processor_df = pd.merge(act_df["_processor_key"].astype(object), processor_df, "right",
                                left_on="_processor_key", right_index=True, )

        # Add a column for function keys by flattening the index of the processor DataFrame.
        processor_df["function_keys"] = processor_df.index.to_flat_index()

        # Remove duplicate rows from the processor DataFrame to ensure uniqueness.
        processor_df = processor_df.drop_duplicates()

        # Add the "process" column to the activity DataFrame using the processor names.
        act_df["process"] = processor_df["name"]

        # Use "product" if available otherwise use "name"
        act_df.update(act_df["product"].rename("name"))
        act_df["product"] = act_df["name"]

        cols = ["amount", "unit", "product", "process", "database", "location", "_processor_key", "_activity_key", "_cs_name"]

        return act_df[cols].reset_index(drop=True)


class FunctionalUnitView(widgets.ABTreeView):
    defaultColumnDelegates = {
        "amount": delegates.AmountDelegate
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.add(actions.ActivityOpen, m.selected_processes,
                               text="Open process" if len(m.selected_processes) == 1 else "Open processes",
                               enable=len(m.selected_processes) > 0
                               ),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(actions.CSDeleteFunctionalUnit, m.cs_name, m.selected_fus,
                               text="Delete Functional Unit" if len(m.selected_fus) == 1 else "Delete Functional Units",
                               enable=len(m.selected_fus) > 0
                               ),

        ]

        @property
        def selected_fus(self):
            return list(set([index.internalPointer().key() for index in self.parent().selectedIndexes()]))

        @property
        def selected_processes(self):
            return list(set([index.internalPointer()["_processor_key"] for index in self.parent().selectedIndexes()]))

        @property
        def cs_name(self):
            return self.parent().parent().calculation_setup_name

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

    def mouseDoubleClickEvent(self, event) -> None:
        """
        Handles the mouse double click event to open the selected activities.

        Args:
            event: The mouse double click event.
        """
        index = self.indexAt(event.pos())
        if index.column() == 0:
            return super().mouseDoubleClickEvent(event)

        if self.selectedIndexes():
            activities = [index.internalPointer()["_processor_key"] for index in self.selectedIndexes()]
            actions.ActivityOpen.run(list(set(activities)))

        return None

    def dragMoveEvent(self, event) -> None:
        pass

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("application/bw-nodekeylist"):
            keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
            for key in keys:
                if not is_node_product(key):
                    keys.remove(key)

            if not keys:
                return

            event.accept()

    def dropEvent(self, event) -> None:
        event.accept()
        cs_name = self.parent().calculation_setup_name

        keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
        for key in keys:
            if not is_node_product(key):
                keys.remove(key)

        actions.CSAddFunctionalUnit.run(cs_name, keys)


class FunctionalUnitItem(widgets.ABDataItem):
    def decorationData(self, col: int, key: str):
        if key == "product":
            return icons.qicons.product
        if key == "process":
            return icons.qicons.process
        return super().decorationData(col, key)

    def flags(self, col: int, key: str):
        """
        Returns the item flags for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the flags.

        Returns:
            QtCore.Qt.ItemFlags: The item flags.
        """
        flags = super().flags(col, key)
        if key in ["amount"]:
            return flags | Qt.ItemFlag.ItemIsEditable
        return flags

    def setData(self, col: int, key: str, value) -> bool:
        """
        Sets the data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to set the data.
            value: The value to set.

        Returns:
            bool: True if the data was set successfully, False otherwise.
        """
        if key not in ["amount"]:
            return False

        cs_name = self["_cs_name"]
        index = self.key()

        actions.CSChangeFunctionalUnit.run(cs_name, index, value)




class FunctionalUnitModel(widgets.ABItemModel):
    dataItemClass = FunctionalUnitItem



from typing import List, Tuple

from peewee import IntegrityError
from qtpy import QtWidgets

from activity_browser import application
from activity_browser.bwutils import refresh_node
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from bw2data.parameters import ActivityParameter
from activity_browser.ui.icons import qicons


class ParameterNewAutomatic(ABAction):
    """
    ABAction for the automatic creation of a new parameter.

    TODO: Remove this action as it is automatic and not user interaction, should be done through e.g. a signal but
    TODO: will actually need to be reworked together with the parameters.
    """

    icon = qicons.add
    text = "New parameter..."

    @staticmethod
    @exception_dialogs
    def run(activities: List[tuple | int | bd.Node]):
        activities = [refresh_node(x) for x in activities]

        for act in activities:
            if act.get("type", "process") not in bd.labels.lci_node_types:
                issue = f"Activity must be 'process' type, '{act.get('name')}' is type '{act.get('type')}'."
                QtWidgets.QMessageBox.warning(
                    application.main_window,
                    "Not allowed",
                    issue,
                    QtWidgets.QMessageBox.Ok,
                    QtWidgets.QMessageBox.Ok,
                )
                return

            group = act.id
            row = {
                "name": "dummy_parameter",
                "amount": act.get("amount", 1.0),
                "formula": act.get("formula", ""),
                "database": act.get("database", ""),
                "code": act.get("code", ""),
            }

            bd.parameters.new_activity_parameters([row], group)



from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class DatabaseProcess(ABAction):
    """
    ...
    """

    icon = qicons.process
    text = "Process database"
    tool_tip = "Process database into datapackages"

    @staticmethod
    @exception_dialogs
    def run(db_name: str):
        db = bd.Database(db_name)
        db.process()



from logging import getLogger

import pandas as pd
from qtpy import QtWidgets, QtCore, QtGui

import bw2data as bd

from activity_browser import signals
from activity_browser.bwutils import AB_metadata
from activity_browser.ui import widgets, application

log = getLogger(__name__)

COLUMNS = ["name", "type", "exchanges", "database", "code"]
DETAILS_COLUMNS = ["input", "output", "type", "amount"]


DEFAULT_STATE = {
    "columns": ["Activity", "Product", "Type", "Unit", "Location"],
    "visible_columns": ["Activity", "Product", "Type", "Unit", "Location"],
}


NODETYPES = {
    "all_nodes": [],
    "processes": ["process", "multifunctional", "processwithreferenceproduct", "nonfunctional"],
    "products": ["product", "processwithreferenceproduct", "waste"],
    "biosphere": ["natural resource", "emission", "inventory indicator", "economic", "social"],
}


class DatabaseExplorerPane(widgets.ABAbstractPane):

    def __init__(self, db_name: str, parent=None):
        super().__init__(parent, QtCore.Qt.WindowType.Window)
        self.title = "Database Explorer - " + db_name
        self.database = bd.Database(db_name)
        self.model = NodeModel(self)

        # Create the QTableView and set the model
        self.table_view = NodeView(self)
        self.table_view.setModel(self.model)
        self.model.setDataFrame(self.build_df())

        self.search = QtWidgets.QLineEdit(self)
        self.search.setMaximumHeight(30)
        self.search.setPlaceholderText("Quick Search")

        self.search.textChanged.connect(self.table_view.setAllFilter)

        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical, self)
        self.splitter.setChildrenCollapsible(False)
        self.splitter.addWidget(self.table_view)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().addWidget(self.search)
        self.layout().addWidget(self.splitter)

        # connect signals
        signals.database.deleted.connect(self.deleteLater)
        signals.project.changed.connect(self.deleteLater)
        AB_metadata.synced.connect(self.sync)
        self.table_view.filtered.connect(self.search_error)

    def sync(self):
        self.model.setDataFrame(self.build_df())

    def build_df(self) -> pd.DataFrame:
        import sqlite3
        from bw2data.backends import sqlite3_lci_db

        full_df = AB_metadata.get_database_metadata(self.database.name)

        con = sqlite3.connect(sqlite3_lci_db._filepath)
        sql = f"SELECT output_code FROM exchangedataset WHERE output_database == '{self.database.name}'"
        excs = pd.read_sql(sql, con)
        con.close()

        count = excs.groupby(excs.columns.tolist()).size()
        count.name = "exchanges"
        full_df = full_df.join(count, "code")

        return full_df

    def search_error(self, reset=False):
        if reset:
            self.search.setPalette(application.palette())
            return

        palette = self.search.palette()
        palette.setColor(QtGui.QPalette.ColorRole.Base, QtGui.QColor(255, 128, 128))
        self.search.setPalette(palette)


class NodeView(widgets.ABTreeView):

    def __init__(self, above: QtWidgets.QWidget=None, parent=None):
        super().__init__(parent)
        self.setSortingEnabled(True)
        self.setDragEnabled(True)
        self.setDragDropMode(QtWidgets.QTableView.DragDropMode.DragOnly)
        self.setSelectionBehavior(widgets.ABTreeView.SelectionBehavior.SelectItems)
        self.setSelectionMode(widgets.ABTreeView.SelectionMode.ExtendedSelection)

        self.above = above
        self.below: QtWidgets.QWidget = QtWidgets.QWidget(self)

    def deleteLater(self):
        super().deleteLater()
        self.below.deleteLater()

    def mouseReleaseEvent(self, event):
        self.below.deleteLater()
        self.below = QtWidgets.QWidget(self)

        if not self.selectedIndexes():
            return

        idx = self.selectedIndexes()[0]
        col_name = self.model().columns()[idx.column()]
        item = idx.internalPointer()
        data = item[col_name]

        if col_name == "exchanges":
            act = bd.get_node(database=item["database"], code=item["code"])
            model = NodeModel()
            model.setDataFrame(pd.DataFrame(act.exchanges()))

            self.below = NodeView(self)
            self.below.setModel(model)

            self.parent().addWidget(self.below)

        elif isinstance(data, (dict, list, tuple)):
            if isinstance(data, dict):
                df = pd.DataFrame.from_dict(data, orient="index")
                df.reset_index(inplace=True)
            else:
                df = pd.DataFrame(data)
            model = NodeModel(dataframe=df)

            self.below = NodeView(self)
            self.below.setModel(model)

            self.parent().addWidget(self.below)

        elif isinstance(data, (str, float, int)):

            if isinstance(data, float) and pd.isna(data):
                return

            self.below = QtWidgets.QPlainTextEdit(str(data), self)
            self.parent().addWidget(self.below)


class NodeItem(widgets.ABDataItem):

    def displayData(self, col: int, key: str):
        data = self[key]

        if data is None:
            return None

        if isinstance(data, (str, float, int)):
            if key == "exchanges":
                return f"Exchanges: {data}" if not pd.isna(data) else "Exchanges: 0"


            rep = str(data).replace("\n", " ")
            if len(rep) > 200:
                return rep[:200] + "..."
            return rep

        elif hasattr(data, "__len__"):
            return f"{type(data).__name__.capitalize()}: {len(data)}"

        else:
            return str(type(data))


class NodeModel(widgets.ABItemModel):
    dataItemClass = NodeItem




from .abstract_pane import ABAbstractPane
from .comparison_switch import SwitchComboBox
from .cutoff_menu import CutoffMenu
from .line_edit import (ABLineEdit, SignalledComboEdit, SignalledLineEdit,
                        SignalledPlainTextEdit)
from .treeview import ABTreeView
from .item_model import ABItemModel
from .item import ABAbstractItem, ABBranchItem, ABDataItem
from .line import ABHLine, ABVLine
from .formula_edit import ABFormulaEdit
from .progress_dialog import ABProgressDialog

from .combobox import ABComboBox
from .button_collapser import ABRadioButtonCollapser
from .wizard import ABWizard
from .wizard_page import ABWizardPage, ABThreadedWizardPage
from .file_selector import ABFileSelector
from .database_name_edit import DatabaseNameEdit
from .dock_widget import ABDockWidget
from .label import ABLabel
from .main_window import MainWindow
from .central import CentralTabWidget
from .menu import ABMenu
from .list_edit_dialog import ABListEditDialog



from bw2analyzer import ContributionAnalysis

from typing import Optional
import numpy as np

class ABContributionAnalysis(ContributionAnalysis):
    """Activity Browser version of bw2analyzer.ContributionAnalysis"""
    def sort_array(self, data: np.array, limit: float = 25, limit_type: str = "number", total: Optional[float] = None) -> np.array:
        """Activity Browser version of bw2analyzer.ContributionAnalysis.sort_array.

        Should be removed once https://github.com/brightway-lca/brightway2-analyzer/pull/32 is merged.
        See PR above on why we overwrite this function.
        """
        if not total:
            total = np.abs(data).sum()

        if total == 0 and limit_type == "cum_percent":
            raise ValueError(
                "Cumulative percentage cannot be calculated to a total of 0, use a different limit type or total")

        if limit_type not in ("number", "percent", "cum_percent"):
            raise ValueError(f"limit_type must be either 'number', 'percent' or 'cum_percent' not '{limit_type}'.")
        if limit_type in ("percent", "cum_percent"):
            if not 0 < limit <= 1:
                raise ValueError("Percentage limits > 0 and <= 1.")
        if limit_type == "number":
            if not int(limit) == limit:
                raise ValueError("Number limit must a whole number.")
            if not 0 < limit:
                raise ValueError("Number limit must be < 0.")

        results = np.hstack(
            (data.reshape((-1, 1)), np.arange(data.shape[0]).reshape((-1, 1)))
        )

        if limit_type == "number":
            # sort and cut off at limit
            return results[np.argsort(np.abs(data))[::-1]][:limit, :]
        elif limit_type == "percent":
            # identify good values, drop rest and sort
            limit = (np.abs(data) >= (abs(total) * limit))
            results = results[limit, :]
            return results[np.argsort(np.abs(results[:, 0]))[::-1]]
        elif limit_type == "cum_percent":
            # if we would apply this on the 'correct' order, this would stop just before the limit,
            # we want to be on or the first step over the limit.
            results = results[np.argsort(np.abs(data))]  # sort low to high impact
            cumsum = np.cumsum(np.abs(results[:, 0])) / abs(total)
            limit = (cumsum >= (1 - limit))  # find items under limit
            return results[limit, :][::-1]  # drop items under limit and set correct order



from qtpy import QtWidgets, QtGui


class ABLabel(QtWidgets.QLabel):
    @classmethod
    def demiBold(cls, *args, **kwargs):
        obj = cls(*args, **kwargs)

        font = obj.font()
        font.setWeight(QtGui.QFont.DemiBold)
        obj.setFont(font)
        return obj



import sys
import os
from logging import getLogger
from importlib import metadata

import requests

from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt

from activity_browser import application
from activity_browser.ui import icons

from .logger import setup_ab_logging
from .static.icons import main

log = getLogger(__name__)


class SpecialProgressBar(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFixedSize(500, 10)

        self.going = "right"

        self.inner = QtWidgets.QFrame(self)
        self.inner.setFixedSize(502, 10)
        self.inner.setStyleSheet("background-color: #0070c0;")
        self.inner.move(100, 0)

        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.loop)
        self.timer.start(1)

    def loop(self):
        x = self.inner.x()

        if self.going == "right":
            x += 1
            if x > 500:
                self.going = "left"
                self.inner.setStyleSheet("background-color: #c00000;")
        elif self.going == "left":
            x -= 1
            if x < -500:
                self.going = "right"
                self.inner.setStyleSheet("background-color: #0070c0;")

        self.inner.move(x, 0)


class ABLoader(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self.setWindowFlags(Qt.FramelessWindowHint)
        self.setWindowTitle("Activity Browser Launcher")
        self.setFixedSize(500, 300)

        logo_pixmap = QtGui.QPixmap(os.path.join(main.__path__[0], "activitybrowser.png")).scaledToHeight(200, mode=Qt.TransformationMode.SmoothTransformation)
        logo_label = QtWidgets.QLabel(self)
        logo_label.setPixmap(logo_pixmap)

        self.text_label = QtWidgets.QLabel("Initializing", self)
        self.text_label.setContentsMargins(0, 0, 30, 0)

        loading_bar = SpecialProgressBar(self)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(logo_label, alignment=Qt.AlignCenter)
        layout.addWidget(self.text_label, alignment=Qt.AlignmentFlag.AlignRight)
        layout.addWidget(loading_bar)
        layout.setContentsMargins(0, 35, 0, 0)

        self.setLayout(layout)

        self.load_modules()

    def load_modules(self):
        thread = ModuleThread(self)
        thread.finished.connect(self.load_layout)
        thread.status.connect(self.text_label.setText)
        thread.start()

    def load_layout(self):
        from .ui.widgets import MainWindow, CentralTabWidget
        from .layouts import panes, pages
        from activity_browser.bwutils import AB_metadata
        from activity_browser import signals

        application.main_window = MainWindow()
        central_widget = CentralTabWidget(application.main_window)
        central_widget.addTab(pages.WelcomePage(), "Welcome")
        central_widget.addTab(pages.ParametersPage(), "Parameters")

        application.main_window.setCentralWidget(central_widget)

        self.load_settings()

    def load_settings(self):
        self.text_label.setText("Loading project")
        thread = SettingsThread(self)
        thread.finished.connect(self.load_finished)
        thread.start()

    def load_finished(self):
        application.main_window.sync()
        application.main_window.show()
        self.deleteLater()


class ModuleThread(QtCore.QThread):
    status: QtCore.SignalInstance = QtCore.Signal(str)

    def run(self):
        self.status.emit("Loading Numpy")
        log.debug("ABLoader: Importing numpy")
        import numpy, pandas
        self.status.emit("Loading Brightway25")
        log.debug("ABLoader: Importing brightway modules")
        import bw2data, bw2calc, bw2analyzer, bw2io, bw_functional, bw_processing, matrix_utils
        self.status.emit("Loading Activity Browser")
        log.debug("ABLoader: Importing activity_browser")
        from activity_browser import actions, layouts, mod, settings, ui, signals
        from activity_browser.layouts import panes, pages
        from activity_browser.ui import core, widgets, web, wizards


class SettingsThread(QtCore.QThread):
    def run(self):
        import bw2data as bd
        from activity_browser import settings, actions

        if settings.ab_settings.settings:
            from pathlib import Path

            base_dir = Path(settings.ab_settings.current_bw_dir)
            project_name = settings.ab_settings.startup_project
            bd.projects.change_base_directories(base_dir, project_name=project_name, update=False)

        if not bd.projects.twofive:
            log.warning(f"Project: {bd.projects.current} is not yet BW25 compatible")
            actions.ProjectSwitch.set_warning_bar()

        log.info(f"Brightway2 data directory: {bd.projects._base_data_dir}")
        log.info(f"Brightway2 current project: {bd.projects.current}")


def run_activity_browser():
    pre_flight_checks()
    setup_ab_logging()
    loader = ABLoader()
    loader.show()
    application.set_icon()  # setting this here seems to fix the icon not showing sometimes
    sys.exit(application.exec_())


def run_activity_browser_no_launcher():
    pre_flight_checks()
    setup_ab_logging()

    modules = ModuleThread()
    modules.run()

    from .ui.widgets import MainWindow, CentralTabWidget
    from .layouts import panes, pages
    from activity_browser.bwutils import AB_metadata
    from activity_browser import signals

    application.main_window = MainWindow()
    central_widget = CentralTabWidget(application.main_window)
    central_widget.addTab(pages.WelcomePage(), "Welcome")
    central_widget.addTab(pages.ParametersPage(), "Parameters")

    application.main_window.setCentralWidget(central_widget)

    settings = SettingsThread()
    settings.run()

    application.main_window.sync()
    application.main_window.show()

    application.set_icon()  # setting this here seems to fix the icon not showing sometimes
    sys.exit(application.exec_())


def pre_flight_checks():

    if "--no-checks" in sys.argv:
        return

    check_pyside_version()

    if "CONDA_DEFAULT_ENV" in os.environ:
        check_conda_update()
    else:
        check_pypi_update()


def check_pyside_version():
    try:
        import PySide6
    except ImportError:
        input("\033[1;31mPySide6 is not installed but highly recommended.\n\n"
              "Please install it using 'pip install PySide6'.\n\n"
              "Press any key to continue...\033[0m")


def check_conda_update():
    ab_url = "https://api.anaconda.org/package/lca/activity-browser"
    try:
        ab_response = requests.get(ab_url)
    except:
        print("Could not fetch latest Activity Browser version")
        return
    ab_current = metadata.version("activity_browser")
    print(f"Activity Browser version: {ab_current}")

    if ab_response.status_code != 200:
        print("Could not fetch latest Activity Browser version")

    elif ab_current != "0.0.0" and ab_current != ab_response.json()['latest_version']:
        input("\033[1;31mThere is an update available for the Activity Browser. Please update it using the following command: \n "
              "conda update -c lca activity-browser\n\n"
              "Press any key to continue without updating...\033[0m")


def check_pypi_update():
    ab_url = "https://pypi.org/pypi/activity-browser/json"
    try:
        ab_response = requests.get(ab_url)
    except:
        print("Could not fetch latest Activity Browser version")
        return
    ab_current = metadata.version("activity_browser")
    print(f"Activity Browser version: {ab_current}")

    if ab_response.status_code != 200:
        print("Could not fetch latest Activity Browser version")

    elif ab_current != "0.0.0" and ab_current != ab_response.json()['info']['version']:
        input("\033[1;31mThere is an update available for the Activity Browser. Please update it using the following command: \n "
              "pip install --upgrade activity-browser\n\n"
              "Press any key to continue without updating...\033[0m")


if "--no-launcher" in sys.argv:
    run_activity_browser_no_launcher()
elif sys.version_info[1] == 10:
    log.info("Running Activity Browser without launcher for Python 3.10")
    run_activity_browser_no_launcher()
else:
    run_activity_browser()



from pathlib import Path

from xlsxwriter.exceptions import FileCreateError

import pandas as pd
import bw2data as bd

from qtpy import QtWidgets, QtCore
from qtpy.QtCore import Qt

from activity_browser import actions, signals
from activity_browser.ui import icons, widgets
from activity_browser.bwutils import manager, superstructure

from .parameter_views import ActivityParameterTable, BaseParameterTable, DataBaseParameterTable, ExchangesTable, ProjectParameterTable, ScenarioTable


class ParametersPage(QtWidgets.QTabWidget):
    """Parameters tab in which user can define project-, database- and
    activity-level parameters for their system.

    Changing projects will trigger a reload of all parameters
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTabsClosable(False)

        # Initialize both parameter tabs
        self.tabs = {
            "Definitions": ParameterDefinitionTab(self),
            "Exchanges": ParameterExchangesTab(self),
            "Scenarios": ParameterScenariosTab(self),
        }
        for name, tab in self.tabs.items():
            self.addTab(tab, name)

        for tab in self.tabs.values():
            if hasattr(tab, "build_tables"):
                tab.build_tables()

        self._connect_signals()

    def _connect_signals(self):
        # signals.add_activity_parameter.connect(self.activity_parameter_added)
        pass

    def activity_parameter_added(self) -> None:
        """Selects the correct sub-tab to show and trigger a switch to
        the Parameters tab.
        """
        self.setCurrentIndex(self.indexOf(self.tabs["Definitions"]))
        signals.show_tab.emit("Parameters")


class ABParameterTable(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.table = None
        self.header = None

    def create_layout(
        self,
        title: str = None,
        bttn: QtWidgets.QAbstractButton = None,
        table: BaseParameterTable = None,
    ):
        headerLayout = QtWidgets.QHBoxLayout()
        self.header = widgets.ABLabel.demiBold(title)

        headerLayout.addWidget(self.header)
        headerLayout.addWidget(bttn)
        headerLayout.addStretch(1)

        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(headerLayout)
        layout.addWidget(table)
        return layout

    def get_table(self):
        return self.table


class ABProjectParameter(ABParameterTable):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.new_parameter_button = actions.ParameterNew.get_QButton(("", ""))
        self.header = "Project:"
        self.table = ProjectParameterTable(self)

        self.setLayout(
            self.create_layout(self.header, self.new_parameter_button, self.table)
        )


class ABDatabaseParameter(ABParameterTable):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.header = "Database:"

        self.new_parameter_button = actions.ParameterNew.get_QButton(("db", ""))

        self.table = DataBaseParameterTable(self)

        self.setLayout(
            self.create_layout(self.header, self.new_parameter_button, self.table)
        )

    def set_enabled(self, trigger):
        if not list(bd.databases):
            self.new_parameter_button.setEnabled(False)
        else:
            self.new_parameter_button.setEnabled(True)


class ABActivityParameter(ABParameterTable):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.header = "Activity:"
        self.parameter = QtWidgets.QCheckBox("Show order column", self)
        self.table = ActivityParameterTable(self)

        self.setLayout(self.create_layout(self.header, self.parameter, self.table))
        self._connect_signal()

    def _connect_signal(self):
        self.parameter.stateChanged.connect(self.activity_order_column)

    def activity_order_column(self) -> None:
        col = self.table.model.order_col
        state = self.parameter.isChecked()
        if not state:
            self.table.setColumnHidden(col, True)
        else:
            self.table.setColumnHidden(col, False)
            self.table.resizeColumnToContents(col)


class ParameterDefinitionTab(QtWidgets.QWidget):
    """Parameter definitions tab.

    This tab shows three tables containing the project-, database- and
    activity level parameters set for the project.

    The user can create new parameters at these three levels and save
    new or edited parameters with a single button.
    Pressing the save button will cause brightway to validate the changes
    and a warning message will appear if an error occurs.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.project_table = ABProjectParameter(self)
        self.database_table = ABDatabaseParameter(self)
        self.activity_table = ABActivityParameter(self)
        self.tables = {
            "project": self.project_table.get_table(),
            "database": self.database_table.get_table(),
            "activity": self.activity_table.get_table(),
        }
        for t in self.tables.values():
            t.model.sync()

        self.show_database_params = QtWidgets.QCheckBox("Database parameters", self)
        self.show_database_params.setToolTip("Show/hide the database parameters")
        self.show_database_params.setChecked(True)

        self.show_activity_params = QtWidgets.QCheckBox("Activity parameters", self)
        self.show_activity_params.setToolTip("Show/hide the activity parameters")
        self.show_activity_params.setChecked(True)
        self.comment_column = QtWidgets.QCheckBox("Comments", self)
        self.comment_column.setToolTip("Show/hide the comment column")
        self.hide_comment_column()
        self.uncertainty_columns = QtWidgets.QCheckBox("Uncertainty", self)
        self.uncertainty_columns.setToolTip("Show/hide the uncertainty columns")

        self._construct_layout()
        self._connect_signals()

        self.explain_text = """
<p>This tab is the main tab for creating and modifying parameters.</p>
<p>The scope of parameters can be either a specific activity, a database, or an entire project 
(meaning that an activity parameter can only be used within a specific activity, 
while a project parameter can be used anywhere within a project and across all databases within that project).</p>



<p><b>In general</b></p>
<p>All parameters must have a <em>name</em> and <em>amount</em>. A <em>formula</em> is optional.</p>
<p>The formula is stored as a string that is interpreted by brightway. Python builtin functions and Numpy functions
can be used within the formula!</p>
<p>Parameters can only be deleted if they are not used in formulas of other parameters.</p>
<p>Note that optionally <a href="https://2.docs.brightway.dev/intro.html#storing-uncertain-values">uncertainties</a>, can be specified for parameters.</p>

<p><b>Activity parameters</b></p>
<p>New parameters are added either by drag-and-dropping activities from the database table or by adding
 a formula to an activity exchange within the Activity tab.</p>
<ul>
<li>Only activities from editable databases can be parameterized.</li>
<li>Multiple parameters can be created for a single activity.</li>
<li>The parameter <em>name</em> must be unique within the group of parameters for an activity.</li>
<li>Note: activity parameters are also auto-generated when a project or database parameter is used in an activity that has previously not been parameterized.</li>
</ul>



<p>For more information on this topic see also the 
<a href="https://2.docs.brightway.dev/intro.html#parameterized-datasets">Brightway2 documentation</a>.</p>
"""

    def _connect_signals(self):
        signals.project.changed.connect(self.build_tables)
        signals.parameter.recalculated.connect(self.build_tables)

        self.show_database_params.toggled.connect(self.hide_database_parameter)
        self.show_activity_params.toggled.connect(self.hide_activity_parameter)
        self.comment_column.stateChanged.connect(self.hide_comment_column)
        self.uncertainty_columns.stateChanged.connect(self.hide_uncertainty_columns)

    def _construct_layout(self):
        """Construct the widget layout for the variable parameters tab"""
        layout = QtWidgets.QVBoxLayout()

        self.uncertainty_columns.setChecked(False)
        row = QtWidgets.QToolBar()
        _header = widgets.ABLabel.demiBold("Parameters ")
        _header.setToolTip("Left click on the question mark for help")
        row.addWidget(_header)
        row.addWidget(self.show_database_params)
        row.addWidget(self.show_activity_params)
        row.addWidget(self.comment_column)
        row.addWidget(self.uncertainty_columns)
        layout.addWidget(row)
        layout.addWidget(widgets.ABHLine(self))

        tables = QtWidgets.QSplitter(Qt.Vertical)
        tables.addWidget(self.project_table)
        tables.addWidget(self.database_table)
        tables.addWidget(self.activity_table)
        layout.addWidget(tables)

        self.setLayout(layout)

    def build_tables(self):
        """Read parameters from brightway and build dataframe tables"""
        self.hide_uncertainty_columns()
        self.activity_order_column()
        # Cannot create database parameters without databases
        if not list(bd.databases):
            self.database_table.set_enabled(False)
        else:
            self.database_table.set_enabled(True)

    def hide_uncertainty_columns(self):
        show = self.uncertainty_columns.isChecked()
        for table in self.tables.values():
            table.uncertainty_columns(show)

    def hide_comment_column(self):
        show = self.comment_column.isChecked()
        for table in self.tables.values():
            table.comment_column(show)

    def activity_order_column(self) -> None:
        col = self.activity_table.get_table().model.order_col
        state = self.activity_table.parameter.isChecked()
        if not state:
            self.activity_table.get_table().setColumnHidden(col, True)
        else:
            self.activity_table.get_table().setColumnHidden(col, False)
            self.activity_table.get_table().resizeColumnToContents(col)

    def hide_database_parameter(self, toggled: bool) -> None:
        self.database_table.header.setHidden(not toggled)
        self.database_table.new_parameter_button.setHidden(not toggled)
        self.database_table.table.setHidden(not toggled)
        self.database_table.setHidden(not toggled)

    def hide_activity_parameter(self, toggled: bool) -> None:
        self.activity_table.header.setHidden(not toggled)
        self.activity_table.parameter.setHidden(not toggled)
        self.activity_table.table.setHidden(not toggled)
        self.activity_table.setHidden(not toggled)


class ParameterExchangesTab(QtWidgets.QWidget):
    """Overview of exchanges

    This tab shows a foldable treeview table containing all of the
    parameters set for the current project.

    Changes made to parameters in the `Definitions` tab will require
    the user to press `Recalculate exchanges` to ensure the amounts in
    the exchanges are properly updated.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.table = ExchangesTable(self)

        self._construct_layout()
        self._connect_signals()

        self.explain_text = """
<p>This tab lists all exchanges within the selected project that are calculated via parameters.</p>
<p>The Project level parameters are shown above the database and activity parameters.</p>
<p>To see the different database and activity parameters in the Project click on the arrows to expand the trees</p>

<p>For more information on this topic see also the 
<a href="https://2.docs.brightway.dev/intro.html#parameterized-datasets">Brightway2 documentation</a>.</p>
"""

    def _connect_signals(self):
        signals.project.changed.connect(self.build_tables)
        signals.parameter.recalculated.connect(self.build_tables)

    def _construct_layout(self):
        """Construct the widget layout for the exchanges parameters tab"""
        layout = QtWidgets.QVBoxLayout()
        row = QtWidgets.QToolBar()
        _header = widgets.ABLabel.demiBold("Overview of parameterized exchanges")
        _header.setToolTip("Left click on the question mark for help")
        row.addWidget(_header)
        row.setIconSize(QtCore.QSize(24, 24))
        layout.addWidget(row)
        layout.addWidget(widgets.ABHLine(self))
        layout.addWidget(self.table, 2)
        self.setLayout(layout)

    def build_tables(self) -> None:
        """Read parameters from brightway and build tree tables"""
        self.table.model.sync()


class ParameterScenariosTab(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.load_btn = QtWidgets.QPushButton(icons.qicons.add, "Import parameter-scenarios")
        self.load_btn.setToolTip(
            "Load prepared excel files with additional parameter scenarios."
        )
        self.save_btn = QtWidgets.QPushButton(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogSaveButton),
            "Export parameter-scenarios",
        )
        self.save_btn.setToolTip(
            "Export the current parameter scenario table to excel."
        )
        self.calculate_btn = QtWidgets.QPushButton(icons.qicons.calculate, "Export as flow-scenarios")
        self.calculate_btn.setToolTip(
            (
                "Process the current parameter scenario table into prepared flow"
                " scenario data."
            )
        )
        self.reset_btn = QtWidgets.QPushButton(icons.qicons.history, "Reset table")
        self.reset_btn.setToolTip("Reset the scenario table, wiping any changes.")
        self.hide_group = QtWidgets.QCheckBox("Show group column")

        self.tbl = ScenarioTable(self)
        self.tbl.setToolTip(
            "This table is not editable, use the export/import functionality"
        )

        self._construct_layout()
        self._connect_signals()

        self.explain_text = """
    <p>This tab has 3 functions:</p>
    <p>1. <b> Export parameter-scenarios </b> : this exports the table as shown below to an Excel file. You can modify it there and use 
    it in scenario LCAs (see Calculation Setup tab)</p>
    <p>2. <b>Import parameter-scenarios</b>: imports a table like the one shown below from Excel. If parameters are missing in Excel, 
    the default values will be used. IMPORTANT NOTE: the ONLY function this button serves is to display the Excel file. 
    If you want to use the Excel file in scenario LCA, please import it in the Calculation Setup tab.</p>
    <p>3. <b>Export as flow-scenarios</b>: This converts a "parameter-scenarios" file (alternative values for parameters) to a 
    "flow-scenarios" file (alternative values for the exchanges as used in LCA calculations).</p>

    <p><b>Suggested <i>workflow</i> to create scenarios for your parameters</b>:</p>
    <p>Export parameter-scenarios. This will generate an Excel file for you where you can add scenarios (columns). 
    You may want to delete rows that you intend to change or rows that are for dependent parameters (those that depend on other parameters) as these values will be overwritten by the formulas. 
    Finally, import the parameter-scenarios in the <i>Calculation Setup</i> (not here!) to perform scenario calculations (you need to select "Scenario LCA").</p>

    <p>For more information on this topic see also the 
    <a href="https://2.docs.brightway.dev/intro.html#parameterized-datasets">Brightway2 documentation</a>.</p>
    """

    def _connect_signals(self):
        self.load_btn.clicked.connect(self.select_read_file)
        self.save_btn.clicked.connect(self.save_scenarios)
        self.calculate_btn.clicked.connect(self.calculate_scenarios)
        self.reset_btn.clicked.connect(self.tbl.model.sync)
        self.hide_group.toggled.connect(self.tbl.group_column)
        signals.parameter_scenario_sync.connect(self.process_scenarios)

    def _construct_layout(self):
        layout = QtWidgets.QVBoxLayout()

        row = QtWidgets.QToolBar()
        _header = widgets.ABLabel.demiBold("Parameter Scenarios")
        _header.setToolTip("Click on the question mark for help")
        row.addWidget(_header)
        layout.addWidget(row)
        layout.addWidget(widgets.ABHLine(self))

        row = QtWidgets.QHBoxLayout()
        # row.addWidget(self.reset_btn)
        row.addWidget(self.save_btn)
        # row.addWidget(self.load_btn)
        row.addWidget(self.calculate_btn)
        # row.addWidget(self.hide_group)
        row.addStretch(1)
        layout.addLayout(row)
        layout.addWidget(self.tbl)
        self.setLayout(layout)

    def process_scenarios(
        self, table_idx: int, df: pd.DataFrame, default: bool
    ) -> None:
        """Use this method to discretely process a parameter scenario file
        for the LCA setup.
        """
        try:
            self.tbl.model.sync(df=df, include_default=default)
            scenarios = self.build_flow_scenarios()
            signals.parameter_superstructure_built.emit(table_idx, scenarios)
        except AssertionError as e:
            QtWidgets.QMessageBox.critical(
                self, "Cannot load parameters", str(e), QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Ok
            )

    def select_read_file(self):
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self, caption="Select prepared scenario file", filter=self.tbl.EXCEL_FILTER
        )
        if path:
            df = pd.read_excel(path, engine="openpyxl")
            self.tbl.model.sync(df=df)

    def save_scenarios(self):
        try:
            self.tbl.to_excel("Save current scenarios to Excel")
        except FileCreateError as e:
            QtWidgets.QMessageBox.warning(
                self,
                "File save error",
                "Cannot save the file, please see if it is opened elsewhere or "
                "if you are allowed to save files in that location:\n\n{}".format(e),
                QtWidgets.QMessageBox.Ok,
                QtWidgets.QMessageBox.Ok,
            )

    def calculate_scenarios(self):
        df = self.build_flow_scenarios()
        self.store_flows_to_file(df)

    def build_flow_scenarios(self) -> pd.DataFrame:
        """Calculate exchange changes for each parameter scenario and construct
        a flow scenarios template file.
        """
        pm = manager.ParameterManager()
        names, data = zip(*self.tbl.iterate_scenarios())

        exchanges = pm.exchanges_from_scenarios(names, data)
        df = superstructure.superstructure_from_scenario_exchanges(exchanges)
        return df

    def store_flows_to_file(self, df: pd.DataFrame) -> None:
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            caption="Save calculated flow scenarios to Excel",
            filter=self.tbl.EXCEL_FILTER,
        )
        if filename:
            try:
                path = Path(filename)
                path = (
                    path
                    if path.suffix in {".xlsx", ".xls"}
                    else path.with_suffix(".xlsx")
                )
                df.to_excel(excel_writer=path, index=False)
            except FileCreateError as e:
                QtWidgets.QMessageBox.warning(
                    self,
                    "File save error",
                    "Cannot save the file, please see if it is opened elsewhere or "
                    "if you are allowed to save files in that location:\n\n{}".format(
                        e
                    ),
                    QtWidgets.QMessageBox.Ok,
                    QtWidgets.QMessageBox.Ok,
                )



from typing import List

from qtpy import QtWidgets

from activity_browser import application, settings
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import commontasks
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

from .activity_open import ActivityOpen


class ActivityDuplicateToDB(ABAction):
    """
    ABAction to duplicate an activity to another database. Asks the user to what database they want to copy the activity
    to, returns if there are no valid databases or when the user cancels. Otherwise uses the activity controller to
    duplicate the activities to the chosen database.
    """

    icon = qicons.duplicate_to_other_database
    text = "Duplicate to other database"

    @classmethod
    @exception_dialogs
    def run(cls, activity_keys: List[tuple], to_db: str = None):
        # from activity_browser.bwutils.metadata import AB_metadata
        activities = [bd.get_activity(key) for key in activity_keys]

        if to_db and not cls.confirm_db(to_db):
            return

        target_db = to_db or cls.request_db(activities)

        if not target_db:
            return

        new_activity_keys = []

        # otherwise move all supplied activities to the db using the controller
        for activity in activities:
            new_code = commontasks.generate_copy_code((target_db, activity["code"]))
            new_activity = activity.copy(code=new_code, database=target_db)
            new_activity_keys.append(new_activity.key)

        ActivityOpen.run(new_activity_keys)

    @staticmethod
    def request_db(activities):
        # get valid databases (not the original database, or locked databases)
        origin_db = next(iter(activities)).get("database")
        target_dbs = [
            db for db in settings.project_settings.get_editable_databases() if db != origin_db
        ]

        # return if there are no valid databases to duplicate to
        if not target_dbs:
            QtWidgets.QMessageBox.warning(
                application.main_window,
                "No target database",
                "No valid target databases available. Create a new database or set one to writable (not read-only).",
            )
            return

        # construct a dialog where the user can choose a database to duplicate to
        target_db, ok = QtWidgets.QInputDialog.getItem(
            application.main_window,
            "Copy activity to database",
            "Target database:",
            target_dbs,
            0,
            False,
        )

        # return if the user didn't choose, or canceled
        if not ok:
            return

        return target_db

    @staticmethod
    def confirm_db(to_db: str):
        user_choice = QtWidgets.QMessageBox.question(
            application.main_window,
            "Duplicate to new database",
            f"Copy to {to_db} and open as new tab?",
        )
        return user_choice == user_choice.Yes



from qtpy import QtWidgets, QtCore, QtGui

import bw2data as bd
import bw_functional as bf

from activity_browser import actions, bwutils, application
from activity_browser.ui import widgets


class ActivityHeader(QtWidgets.QWidget):
    """
    A widget that displays the header information of a specific activity.

    Attributes:
        DATABASE_DEFINED_ALLOCATION (str): Constant for database default allocation.
        CUSTOM_ALLOCATION (str): Constant for custom allocation.
        activity (bd.Node): The activity to display the header for.
    """
    DATABASE_DEFINED_ALLOCATION = "(database default)"
    CUSTOM_ALLOCATION = "Custom..."

    def __init__(self, parent: QtWidgets.QWidget):
        """
        Initializes the ActivityHeader widget.

        Args:
            parent (QtWidgets.QWidget): The parent widget.
        """
        super().__init__(parent)
        self.activity = parent.activity

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def sync(self):
        """
        Synchronizes the widget with the current state of the activity.
        """
        self.activity = bwutils.refresh_node(self.activity)

        self.clear_layout()

        if bwutils.database_is_locked(self.activity["database"]):
            self.layout().addWidget(LockedWarningBar(self))

        self.layout().addLayout(self.build_grid())

    def clear_layout(self, layout: QtWidgets.QLayout = None):
        layout = layout or self.layout()

        if layout is None:
            return

        while layout.count():
            item = layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()
            elif item.layout() is not None:
                self.clear_layout(item.layout())

    def build_grid(self) -> QtWidgets.QGridLayout:
        grid = QtWidgets.QGridLayout(self)
        grid.setContentsMargins(0, 5, 0, 5)
        grid.setSpacing(10)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        db_locked = bwutils.database_is_locked(self.activity["database"])
        setup = self.disabled_setup() if db_locked else self.enabled_setup()

        # Arrange widgets for display as a grid
        for i, (title, widget) in enumerate(setup):
            grid.addWidget(widgets.ABLabel.demiBold(title, self), i, 1)
            grid.addWidget(widget, i, 2, 1, 4)

        return grid

    def enabled_setup(self):
        setup = [
            ("Name:", ActivityName(self),),
            ("Location:", ActivityLocation(self),),
        ]

        if isinstance(self.activity, bf.Process):
            setup.append(("Properties:", ActivityProperties(self),),)

        # Add allocation strategy selector if the activity is multifunctional
        if self.activity.get("type") == "multifunctional":
            setup.append(("Allocation:", ActivityAllocation(self),))

        return setup

    def disabled_setup(self):
        setup = [
            ("Name:", QtWidgets.QLabel(self.activity.get("name", "unspecified"), self),),
            ("Location:", QtWidgets.QLabel(self.activity.get("location", "unspecified"), self),),
        ]

        if isinstance(self.activity, bf.Process):
            props = self.activity.available_properties()
            prop_text = ", ".join(props) if props else "None"
            setup.append(("Properties:", QtWidgets.QLabel(prop_text, self),))

        # Add allocation strategy selector if the activity is multifunctional
        if self.activity.get("type") == "multifunctional":
            setup.append(("Allocation:", QtWidgets.QLabel(self.activity.get("allocation", "unspecified"), self),),)

        return setup





class ActivityName(QtWidgets.QLineEdit):
    """
    A widget that displays and edits the name of the activity.
    """

    def __init__(self, parent: ActivityHeader):
        """
        Initializes the ActivityName widget.

        Args:
            parent (ActivityHeader): The parent widget.
        """
        super().__init__(parent.activity["name"], parent)
        self.editingFinished.connect(self.change_name)

    def change_name(self):
        """
        Changes the name of the activity if it has been modified.
        """
        if self.text() == self.parent().activity["name"]:
            return
        actions.ActivityModify.run(self.parent().activity, "name", self.text())


class ActivityLocation(QtWidgets.QLineEdit):
    """
    A widget that displays and edits the location of the activity.
    """

    def __init__(self, parent: ActivityHeader):
        """
        Initializes the ActivityLocation widget.

        Args:
            parent (ActivityHeader): The parent widget.
        """
        super().__init__(parent.activity.get("location"), parent)
        self.editingFinished.connect(self.change_location)

        locations = set(bwutils.AB_metadata.dataframe.get("location", ["GLO"]))
        completer = QtWidgets.QCompleter(locations, self)
        self.setCompleter(completer)

    def change_location(self):
        """
        Changes the location of the activity if it has been modified.
        """
        if self.text() == self.parent().activity.get("location"):
            return
        actions.ActivityModify.run(self.parent().activity, "location", self.text())


class ActivityProperties(QtWidgets.QWidget):
    """
    A widget that displays and edits the properties of the activity.
    """

    def __init__(self, parent: ActivityHeader):
        """
        Initializes the ActivityProperties widget.

        Args:
            parent (ActivityHeader): The parent widget.
        """
        super().__init__(parent)

        self.setContentsMargins(0, 0, 0, 0)
        layout = QtWidgets.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        if not isinstance(parent.activity, bf.Process):
            return

        for property_name in parent.activity.available_properties():
            layout.addWidget(ActivityProperty(parent.activity, property_name))

        add_label = QtWidgets.QLabel("<a style='text-decoration:underline;'>Add property</a>")
        add_label.mouseReleaseEvent = lambda x: actions.ProcessPropertyModify.run(parent.activity)

        layout.addWidget(add_label)

        layout.addStretch(1)


class ActivityProperty(QtWidgets.QPushButton):
    """
    A widget that represents a single property of the activity.
    """

    def __init__(self, activity, property_name):
        """
        Initializes the ActivityProperty widget.

        Args:
            activity (bd.Node): The activity to which the property belongs.
            property_name (str): The name of the property.
        """
        super().__init__(property_name, None)

        self.modify_action = actions.ProcessPropertyModify.get_QAction(activity, property_name)
        self.remove_action = actions.ProcessPropertyRemove.get_QAction(activity, property_name)

        self.menu = QtWidgets.QMenu(self)
        self.menu.addAction(self.modify_action)
        self.menu.addAction(self.remove_action)

        self.setStyleSheet("""
        QPushButton {
            border: 1px solid #8f8f91;
            border-radius: 0px;
            padding: 1px 10px 1px 10px;
            min-width: 0px;
        }
        """)

    def mouseReleaseEvent(self, e):
        """
        Handles the mouse release event to show the context menu.

        Args:
            e: The mouse release event.
        """
        pos = self.geometry().bottomLeft()
        pos = self.parent().mapToGlobal(pos)
        self.menu.exec_(pos)
        e.accept()


class ActivityAllocation(QtWidgets.QComboBox):
    """
    A widget that displays and edits the allocation strategy of the activity.
    """

    def __init__(self, parent: ActivityHeader):
        """
        Initializes the ActivityAllocation widget.

        Args:
            parent (ActivityHeader): The parent widget.
        """
        if not isinstance(parent.activity, bf.Process):
            raise TypeError("ActivityAllocation can only be used with bf.Process instances.")

        super().__init__(parent)

        self.addItems(sorted(bf.allocation_strategies))
        if props := parent.activity.available_properties():
            self.insertSeparator(1000)  # Large number to make sure it's appended at the end
            self.addItems(sorted(props))

        i = self.findText(parent.activity.get("allocation"))
        self.setCurrentIndex(i)

        self.currentTextChanged.connect(self.change_allocation)

    def change_allocation(self, allocation: str):
        """
        Changes the allocation strategy of the activity if it has been modified.

        Args:
            allocation (str): The new allocation strategy.
        """
        act = self.parent().activity
        if act.get("allocation") == allocation:
            return
        actions.ActivityModify.run(act, "allocation", allocation)


class LockedWarningBar(QtWidgets.QToolBar):
    def __init__(self, parent: ActivityHeader):
        super().__init__(parent)
        self.setMovable(False)
        self.setContentsMargins(0, 0, 0, 0)

        warning_label = QtWidgets.QLabel("The database of this activity is currently locked.")
        height = warning_label.minimumSizeHint().height()

        warning_icon = QtWidgets.QLabel(self)
        qicon = application.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_MessageBoxWarning)
        pixmap = qicon.pixmap(height, height)
        warning_icon.setPixmap(pixmap)

        migrate_label = QtWidgets.QLabel("<a style='text-decoration:underline;'>Unlock database</a>")
        migrate_label.mouseReleaseEvent = lambda x: actions.DatabaseSetReadonly.run(parent.activity["database"], False)

        self.addWidget(warning_icon)
        self.addWidget(warning_label)
        self.addWidget(migrate_label)

    def contextMenuEvent(self, event):
        return None





from typing import Optional, Tuple
from qtpy.QtWidgets import QDialog, QGridLayout, QLabel, QLineEdit, QPushButton, QWidget


class NewNodeDialog(QDialog):
    """
    Gathers the paremeters for creating a new process.
    """

    def __init__(self, process: bool = True, parent: Optional[QWidget] = None):
        super().__init__(parent)
        layout = QGridLayout()
        row = 0
        if process:
            self.setWindowTitle("New process")
            layout.addWidget(QLabel("Process name"), row, 0)
        else:
            self.setWindowTitle("New product")
            layout.addWidget(QLabel("Product name"), row, 0)
        self._process_name_edit = QLineEdit()
        self._process_name_edit.textChanged.connect(self._handle_text_changed)
        layout.addWidget(self._process_name_edit, row, 1)
        row += 1
        self._ref_product_name_edit = QLineEdit()
        if process:
            layout.addWidget(QLabel("Product name"), row, 0)
            layout.addWidget(self._ref_product_name_edit, row, 1)
            row += 1
        layout.addWidget(QLabel("Unit"), row, 0)
        self._unit_edit = QLineEdit("kilogram")
        layout.addWidget(self._unit_edit, row, 1)
        row += 1
        layout.addWidget(QLabel("Location"), row, 0)
        default_loc = "GLO" if process else ""
        self._location_edit = QLineEdit(default_loc)
        layout.addWidget(self._location_edit, row, 1)
        row += 1
        self._ok_button = QPushButton("OK")
        self._ok_button.clicked.connect(self.accept)
        self._ok_button.setEnabled(False)
        layout.addWidget(self._ok_button, row, 0)
        cancel_button = QPushButton("Cancel")
        cancel_button.clicked.connect(self.reject)
        layout.addWidget(cancel_button, row, 1)
        self.setLayout(layout)

    def _handle_text_changed(self, text: str):
        self._ok_button.setEnabled(text != "")
        self._ref_product_name_edit.setPlaceholderText(text)

    def get_new_process_data(self) -> Tuple[str, str, str, str]:
        """Return the parameters the user entered."""
        return (
                self._process_name_edit.text(),
                self._ref_product_name_edit.text(),
                self._unit_edit.text(),
                self._location_edit.text()
            )





import re

from qtpy import QtWidgets

from .dock_widget import ABDockWidget


class ABAbstractPane(QtWidgets.QWidget):
    title: str
    name: str = None
    unique: bool = False  # whether the pane is unique in the application

    def __init__(self, parent=None):
        super().__init__(parent)
        self.name = self.name or re.sub(r'([a-z])([A-Z])', r'\1_\2', self.__class__.__name__).lower()
        self.setObjectName(self.name)
        self.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding)

    def getDockWidget(self, main_window: QtWidgets.QMainWindow):

        hidemode = ABDockWidget.HideMode.Hide if self.unique else ABDockWidget.HideMode.Close

        dock_widget = ABDockWidget(self.title, parent=main_window, mode=hidemode)
        dock_widget.setWidget(self)
        return dock_widget

    def sync(self):
        """
        Synchronize the pane with the current state of Brightway.
        """
        pass

    def saveState(self):
        """
        Save the state of the pane.
        """
        return {}

    def restoreState(self, state: dict):
        """
        Restore the state of the pane.
        """
        pass

    @classmethod
    def fromState(cls, state: dict, parent=None):
        """
        Restore the state of the pane.
        """
        pane = cls(parent=parent)
        pane.restoreState(state)
        return pane



import numpy as np
import pandas as pd

import bw2data as bd
from bw2data.backends import ActivityDataset


FROM_ACT = pd.Index(
    ["from activity name", "from reference product", "from location", "from database"]
)
TO_ACT = pd.Index(
    [
        "to activity name",
        "to reference product",
        "to location",
        "to database",
    ]
)
FROM_BIOS = pd.Index(["from activity name", "from categories", "from database"])
TO_BIOS = pd.Index(["to activity name", "to categories", "to database"])
FROM_ALL = pd.Index(
    [
        "from activity name",
        "from reference product",
        "from location",
        "from categories",
        "from database",
        "from key",
    ]
)
TO_ALL = pd.Index(
    [
        "to activity name",
        "to reference product",
        "to location",
        "to categories",
        "to database",
        "to key",
    ]
)


def process_ad_namedtuple(row) -> tuple:
    """Take a given ActivityDataset namedtuple and return two hashable tuples.

    Allows for matching on name/product/location
    """
    match = (row.name, row.product, row.location)
    key = (row.database, row.code)
    return match, key


def process_ad_flow(row) -> tuple:
    match = (row.name, row.data.get("categories", None))
    key = (row.database, row.code)
    return match, key


def construct_ad_data(row) -> tuple:
    """Take a namedtuple from the method below and convert it into two tuples.

    Used to fill out missing information in the superstructure.
    """
    key = (row.database, row.code)
    if row.type == "process":
        data = (row.name, row.product, row.location, np.NaN, row.database)
    elif "categories" in row.data:
        data = (row.name, np.NaN, np.NaN, row.data["categories"], row.database)
    else:
        data = (row.name, np.NaN, np.NaN, np.NaN, row.database)
    return key, data


def data_from_index(index: tuple) -> dict:
    """Take the given 'Index' tuple and build a complete SUPERSTRUCTURE row
    from it.
    """
    from_key, to_key = index[0], index[1]
    from_key, from_data = construct_ad_data(
        ActivityDataset.get(database=from_key[0], code=from_key[1])
    )
    to_key, to_data = construct_ad_data(
        ActivityDataset.get(database=to_key[0], code=to_key[1])
    )
    return {
        "from activity name": from_data[0],
        "from reference product": from_data[1],
        "from location": from_data[2],
        "from categories": from_data[3],
        "from database": from_data[4],
        "from key": from_key,
        "to activity name": to_data[0],
        "to reference product": to_data[1],
        "to location": to_data[2],
        "to categories": to_data[3],
        "to database": to_data[4],
        "to key": to_key,
        "flow type": index[2] if len(index) > 2 else np.NaN,
    }


def get_relevant_activities(df: pd.DataFrame, part: str = "from") -> dict:
    """Build a dictionary of (name, product, location) -> (database, key) pairs."""
    select = FROM_ACT if part == "from" else TO_ACT
    sub = df.loc[:, select]
    sub = sub[sub.iloc[:, 3] != bd.config.biosphere]  # Exclude biosphere exchanges
    if sub.empty:
        return {}

    names, products, locations, dbs = list(map(set, sub.iloc[:, 0:4].values.T))
    #    names, products, locations, dbs = sub.iloc[:, 0:4].apply(set, axis=0)
    query = (
        ActivityDataset.select(
            ActivityDataset.name,
            ActivityDataset.product,
            ActivityDataset.location,
            ActivityDataset.database,
            ActivityDataset.code,
        )
        .where(
            (ActivityDataset.name.in_(names))
            & (ActivityDataset.product.in_(products))
            & (ActivityDataset.location.in_(locations))
            & (ActivityDataset.database.in_(dbs))
        )
        .namedtuples()
    )
    activities = dict(map(process_ad_namedtuple, query.iterator()))
    return activities


def get_relevant_flows(df: pd.DataFrame, part: str = "from") -> dict:
    """Determines if all activities from the given 'from' or 'to' chunk"""
    select = FROM_BIOS if part == "from" else TO_BIOS
    sub = df.loc[:, select]
    sub = sub[sub.iloc[:, 2] == bd.config.biosphere]  # Use only biosphere exchanges
    if sub.empty:
        return {}

    names, categories, dbs = list(map(set, sub.iloc[:, 0:3].values.T))
    #    names, categories, dbs = sub.iloc[:, 0:3].apply(set, axis=0)
    query = (
        ActivityDataset.select(
            ActivityDataset.name,
            ActivityDataset.data,
            ActivityDataset.database,
            ActivityDataset.code,
        )
        .where((ActivityDataset.name.in_(names)) & (ActivityDataset.database.in_(dbs)))
        .namedtuples()
    )
    flows = dict(map(process_ad_flow, query.iterator()))
    return flows


def match_fields_for_key(df: pd.DataFrame, matchbook: dict) -> pd.Series:
    def build_match(row):
        if row.iat[4] == bd.config.biosphere:
            match = (row.iat[0], row.iat[3])
        else:
            match = (row.iat[0], row.iat[1], row.iat[2])
        return matchbook.get(match, np.NaN)

    return df.apply(build_match, axis=1)


def fill_df_keys_with_fields(df: pd.DataFrame) -> pd.DataFrame:
    matches = get_relevant_flows(df, "from")
    matches.update(get_relevant_activities(df, "from"))
    df["from key"] = match_fields_for_key(df.loc[:, FROM_ALL], matches)
    matches = get_relevant_flows(df, "to")
    matches.update(get_relevant_activities(df, "to"))
    df["to key"] = match_fields_for_key(df.loc[:, TO_ALL], matches)
    return df


def get_activities_from_keys(
    df: pd.DataFrame, db: str = bd.config.biosphere
) -> pd.DataFrame:
    """
    Uses the BW SQL database to generate a list of Activities from the input dataframe.
    Returns a pandas dataframe that contains any keys that do not identify to an Activity in BW.

    parameters
    ----------
    df: pandas dataframe for a scenario

    db: the database name to check the Activities from the dataframe to
    """
    data_f = df.loc[(df["from database"] == db)]
    data_t = df.loc[(df["to database"] == db)]
    flows = set()
    if not data_f.empty:
        f_db, f_keys = zip(
            *data_f.loc[:, "from key"]
        )  # extract just the key, avoiding the database
        fqry = (
            ActivityDataset.select(ActivityDataset.code, ActivityDataset.database)
            .where(
                (ActivityDataset.database == db)
                & (ActivityDataset.code.in_(set(f_keys)))
            )
            .namedtuples()
        )  # produces an iterator for the activities
        flows.update(set(map(lambda row: (row.database, row.code), fqry.iterator())))
    if not data_t.empty:
        t_db, t_keys = zip(*data_t.loc[:, "to key"])  # look at the above code block
        tqry = (
            ActivityDataset.select(ActivityDataset.code, ActivityDataset.database)
            .where(
                (ActivityDataset.database == db)
                & (ActivityDataset.code.in_(set(t_keys)))
            )
            .namedtuples()
        )

        flows.update(set(map(lambda row: (row.database, row.code), tqry.iterator())))
    absent = pd.concat(
        [
            data_f.loc[
                ~(data_f["from key"].isin(flows)) & (data_f["from database"] == db)
            ],
            data_t.loc[~(data_t["to key"].isin(flows)) & (data_t["to database"] == db)],
        ],
        ignore_index=False,
        axis=0,
    )
    # absent includes those exchanges where one of the keys was not found in the respective database
    return absent



from typing import List

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class CFRemove(ABAction):
    """
    ABAction to remove one or more Characterization Factors from a method. First ask for confirmation and return if the
    user cancels. Otherwise instruct the ImpactCategoryController to remove the selected Characterization Factors.
    """

    icon = qicons.delete
    text = "Remove CF('s)"

    @staticmethod
    @exception_dialogs
    def run(method_name: tuple, char_factors: List[tuple]):
        # ask the user whether they are sure to delete the calculation setup
        warning = QtWidgets.QMessageBox.warning(
            application.main_window,
            "Deleting Characterization Factors",
            f"Are you sure you want to delete {len(char_factors)} CF('s)?",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No,
        )

        # return if the users cancels
        if warning == QtWidgets.QMessageBox.No:
            return

        method = bd.Method(method_name)
        method_dict = {cf[0]: cf[1] for cf in method.load()}

        for cf in char_factors:
            method_dict.pop(cf[0])

        method.write(list(method_dict.items()))



from qtpy import QtGui
from logging import getLogger

from activity_browser import signals
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd

log = getLogger(__name__)

class MultifunctionalProcessRedoAllocation(ABAction):
    """
    ABAction to redo the allocation calculation for a specific process.
    """

    icon = QtGui.QIcon()
    text = "Redo allocation for multifunctional process"
    tool_tip = "Redo the allocation calculations for this process"

    @staticmethod
    @exception_dialogs
    def run(node: bd.Node):
        if not getattr(node, "multifunctional", None):
            return
        try:
            node.allocate()

            signals.new_statusbar_message.emit(f"Allocation values for process {node} updated.")
        except KeyError as exc:
            signals.new_statusbar_message.emit("A property for the allocation calculation was not found!")
            log.error(f"A property for the allocation calculation was not found: {node}")
            raise exc
        except ZeroDivisionError as exc:
            signals.new_statusbar_message.emit(str(exc))
            log.error(f"Zero division in allocation calculation: {exc}")
            raise exc



# -*- coding: utf-8 -*-
from os import devnull

from asteval import Interpreter
from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import Signal, Slot

from activity_browser import actions, signals


class CalculatorButtons(QtWidgets.QWidget):
    """A custom layout containing calculator buttons, emits a signal
    for each button pressed.
    """

    button_press = Signal(str)
    clear = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.explain_text = """
In addition to the other buttons on this calculator, the parameter formula
can make use of a large number of Python and Numpy functions, with Numpy
overriding Python where the function names are the same.

For a more complete list see the `math` module in the Python documentation
or `ufuncs` in de Numpy documentation.

Keep in mind that the result of a formula must be a scalar value!
"""
        rows = [
            [
                ("+", "Add", lambda: self.button_press.emit(" + ")),
                ("-", "Subtract", lambda: self.button_press.emit(" - ")),
                ("*", "Multiply", lambda: self.button_press.emit(" * ")),
            ],
            [
                ("/", "Divide", lambda: self.button_press.emit(" / ")),
                ("x²", "X to the power of 2", lambda: self.button_press.emit(" ** 2 ")),
                ("More...", "Additional functions", self.explanation),
            ],
        ]
        # Construct the layout from the list of lists above.
        layout = QtWidgets.QHBoxLayout()
        layout.addStretch(1)
        for row in rows:
            bar = QtWidgets.QToolBar()
            bar.setOrientation(QtCore.Qt.Vertical)
            for btn in row:
                w = QtWidgets.QPushButton(btn[0])
                w.setToolTip(btn[1])
                w.pressed.connect(btn[2])
                w.setFixedSize(50, 50)
                bar.addWidget(w)
            layout.addWidget(bar)
        layout.addStretch(1)
        self.setLayout(layout)

    @Slot()
    def explanation(self):
        return QtWidgets.QMessageBox.question(
            self,
            "More...",
            self.explain_text,
            QtWidgets.QMessageBox.Ok,
            QtWidgets.QMessageBox.Ok,
        )


class FormulaDialog(QtWidgets.QDialog):
    def __init__(self, parent=None, flags=QtCore.Qt.Window):
        super().__init__(parent=parent, f=flags)
        self.setWindowTitle("Build a formula")
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.interpreter = None
        self.key = ("", "")

        # 6 broad by 6 deep.
        grid = QtWidgets.QGridLayout(self)
        self.text_field = QtWidgets.QLineEdit(self)
        self.text_field.textChanged.connect(self.validate_formula)
        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Save | QtWidgets.QDialogButtonBox.Cancel
        )
        self.buttons.setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Expanding,
                QtWidgets.QSizePolicy.Preferred,
                QtWidgets.QSizePolicy.ButtonBox,
            )
        )
        self.parameters = QtWidgets.QTableView(self)
        model = QtGui.QStandardItemModel(self)
        self.parameters.setModel(model)
        completer = QtWidgets.QCompleter(model, self)
        completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.text_field.setCompleter(completer)
        self.parameters.doubleClicked.connect(self.append_parameter_name)

        self.new_parameter_button = actions.ParameterNew.get_QButton(self.get_key)

        self.calculator = CalculatorButtons(self)
        self.calculator.button_press.connect(self.text_field.insert)
        self.calculator.clear.connect(self.text_field.clear)

        grid.addWidget(self.text_field, 0, 0, 5, 1)
        grid.addWidget(self.buttons, 5, 0, 1, 1)
        grid.addWidget(self.calculator, 0, 1, 5, 1)
        grid.addWidget(self.parameters, 0, 2, 5, 1)
        grid.addWidget(self.new_parameter_button, 5, 2, 1, 1)

        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)
        signals.added_parameter.connect(self.append_parameter)
        self.show()

    def insert_parameters(self, items) -> None:
        """Take the given list of parameter names, amounts and types, insert
        them into the model.
        """
        model = self.parameters.model()
        model.clear()
        model.setHorizontalHeaderLabels(["Name", "Amount", "Type"])
        for x, item in enumerate(items):
            for y, value in enumerate(item):
                model_item = QtGui.QStandardItem(str(value))
                model_item.setEditable(False)
                model.setItem(x, y, model_item)
        self.parameters.resizeColumnsToContents()

    @Slot(str, str, str, name="appendParameter")
    def append_parameter(self, name: str, amount: str, p_type: str) -> None:
        """Catch new parameters from the wizard and add them to the list."""
        model = self.parameters.model()
        x = model.rowCount()
        for y, i in enumerate([name, amount, p_type]):
            item = QtGui.QStandardItem(i)
            item.setEditable(False)
            model.setItem(x, y, item)

        # Also include the new parameter in the interpreter.
        if self.interpreter:
            self.interpreter.symtable.update({name: float(amount)})

    def insert_interpreter(self, interpreter: Interpreter) -> None:
        self.interpreter = interpreter

    def insert_key(self, key: tuple) -> None:
        """The key consists of two strings, no more, no less."""
        self.key = key

    def get_key(self) -> tuple:
        return self.key

    @property
    def formula(self) -> str:
        """Look into the text_field and return the formula."""
        return self.text_field.text().strip()

    @formula.setter
    def formula(self, value) -> None:
        """Take the formula and set it to the text_field widget."""
        if value is None:
            self.text_field.clear()
        else:
            self.text_field.setText(str(value))

    @Slot(QtCore.QModelIndex)
    def append_parameter_name(self, index: QtCore.QModelIndex) -> None:
        """Take the index from the parameters table and append the parameter
        name to the formula.
        """
        param_name = self.parameters.model().index(index.row(), 0).data()
        self.text_field.insert(param_name)

    @Slot()
    def validate_formula(self) -> None:
        """Qt slot triggered whenever a change is detected in the text_field."""
        self.text_field.blockSignals(True)
        if self.interpreter:
            formula = self.text_field.text().strip()
            # Do not write massive amounts of errors to stderr if the user
            # is busy writing.
            with open(devnull, "w") as errfile:
                self.interpreter.err_writer = errfile
                self.interpreter(formula)
                if len(self.interpreter.error) > 0:
                    self.buttons.button(QtWidgets.QDialogButtonBox.Save).setEnabled(
                        False
                    )
                else:
                    self.buttons.button(QtWidgets.QDialogButtonBox.Save).setEnabled(
                        True
                    )
        self.text_field.blockSignals(False)


class FormulaDelegate(QtWidgets.QStyledItemDelegate):
    """An extensive delegate to allow users to build and validate formulas
    The delegate spawns a dialog containing:
      - An editable textfield for the formula.
      - A listview containing parameter names that can be used in the formula
      - Ok and Cancel buttons, on Ok, validate the formula before saving
    For hardmode: also allow the user to create a new parameter from WITHIN
    the delegate dialog itself. Requiring us to also include refreshing
    for the parameter list.
    """

    ACCEPTED_TABLES = {
        "project_parameter",
        "database_parameter",
        "activity_parameter",
        "product",
        "technosphere",
        "biosphere",
    }

    def __init__(self, parent=None):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        editor = QtWidgets.QWidget(parent)
        dialog = FormulaDialog(editor, QtCore.Qt.Window)
        dialog.accepted.connect(lambda: self.commitData.emit(editor))
        # dialog.rejected.connect(signals.parameters_changed.emit)
        return editor

    def setEditorData(self, editor: QtWidgets.QWidget, index: QtCore.QModelIndex):
        """Populate the editor with data if editing an existing field."""
        dialog = editor.findChild(FormulaDialog)
        data = index.data(QtCore.Qt.DisplayRole)

        parent = self.parent()
        # Check which table is asking for a list
        if getattr(parent, "table_name", "") in self.ACCEPTED_TABLES:
            items = parent.get_usable_parameters()
            dialog.insert_parameters(items)
            dialog.formula = data
            interpreter = parent.get_interpreter()
            dialog.insert_interpreter(interpreter)
            # Now see if we can construct a (partial) key
            if hasattr(parent, "key"):
                # This works for exchange tables.
                dialog.insert_key(parent.key)
            elif hasattr(parent, "get_key"):
                dialog.insert_key(parent.get_key())

    def setModelData(
        self,
        editor: QtWidgets.QWidget,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ):
        """Take the editor, read the given value and set it in the model.

        If the new formula is the same as the existing one, do not call setData
        """
        dialog = editor.findChild(FormulaDialog)
        if dialog.result() == QtWidgets.QDialog.Rejected:
            # Cancel was clicked, do not store anything.
            return
        model.setData(index, dialog.formula, QtCore.Qt.EditRole)



import sys

# This patch fixes a bug on Python 3.10 where shiboken6 patches the typing module with a broken Self type. Only way to
# fix it is to patch it ourselves first.
if sys.version_info[1] < 11:
    import typing
    if hasattr(typing, "Self"): # if Self has already been patched into typing, delete it first
        del typing.Self
    import typing_extensions
    setattr(typing, "Self", typing_extensions.Self)

try:
    import PySide6
except ImportError:
    import qtpy

from .ui.application import application
from .signals import signals

def run_activity_browser():
    from .__main__ import run_activity_browser



"""Custom page for debugging javascript code. Without this code,
    only console.error messages are printed to python output.
    This code will not tell you the javascript file that the error is in."""
from logging import getLogger

from qtpy.QtWebEngineWidgets import QWebEnginePage

log = getLogger(__name__)


class Page(QWebEnginePage):
    def javaScriptConsoleMessage(self, level: QWebEnginePage.JavaScriptConsoleMessageLevel, message: str, line: str, _: str):
        if level == QWebEnginePage.InfoMessageLevel:
            log.info(f"JS Info (Line {line}): {message}")
        elif level == QWebEnginePage.WarningMessageLevel:
            log.warning(f"JS Warning (Line {line}): {message}")
        elif level == QWebEnginePage.ErrorMessageLevel:
            log.error(f"JS Error (Line {line}): {message}")
        else:
            log.debug(f"JS Log (Line {line}): {message}")



from qtpy import QtCore, QtGui, QtWidgets

class AmountDelegate(QtWidgets.QStyledItemDelegate):
    def displayText(self, value, locale):
        import math
        try:
            value = float(value)
        except ValueError:
            value = math.nan

        if math.isnan(value):
            return ""
        return str(abs(value))

    def createEditor(self, parent, option, index):
        editor = QtWidgets.QLineEdit(parent)
        locale = QtCore.QLocale(QtCore.QLocale.English)
        locale.setNumberOptions(QtCore.QLocale.RejectGroupSeparator)
        validator = QtGui.QRegularExpressionValidator(
            QtCore.QRegularExpression(r"^[+-]?((\d+(\.\d*)?)|(\.\d+))([eE][+-]?\d+)?$"),
            editor)
        validator.setLocale(locale)
        editor.setValidator(validator)
        return editor

    def setEditorData(self, editor: QtWidgets.QLineEdit, index: QtCore.QModelIndex):
        """Populate the editor with data if editing an existing field."""
        import math

        data = index.data(QtCore.Qt.DisplayRole)

        try:
            value = abs(float(data))
        except ValueError:
            value = math.nan

        editor.setText(format(value, '.10f').rstrip('0').rstrip('.'))

    def setModelData(
            self,
            editor: QtWidgets.QLineEdit,
            model: QtCore.QAbstractItemModel,
            index: QtCore.QModelIndex,
    ):
        """Take the editor, read the given value and set it in the model"""
        try:
            value = float(editor.text())
            old = float(index.data(QtCore.Qt.DisplayRole))

            if old < 0:
                value = value * -1

            model.setData(index, value, QtCore.Qt.EditRole)
        except ValueError:
            pass


from qtpy import QtWidgets

from activity_browser.ui import widgets
from activity_browser.bwutils import AB_metadata


class MetaDataStorePage(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName("MetaDataStorePage")

        self.model = MDSModel(self, AB_metadata.dataframe)
        self.view = MDSView(self)
        self.view.setModel(self.model)

        self.build_layout()
        self.connect_signals()

    def connect_signals(self):
        AB_metadata.synced.connect(self.sync)

    def sync(self):
        self.model.setDataFrame(AB_metadata.dataframe)

    def build_layout(self):
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        self.setLayout(layout)


class MDSView(widgets.ABTreeView):
    pass


class MDSItem(widgets.ABDataItem):
    pass

class MDSModel(widgets.ABItemModel):
    pass



from qtpy import QtWidgets, QtCore


class PropertyDelegate(QtWidgets.QStyledItemDelegate):
    """
    A delegate for displaying and editing property values in the tree view.
    """

    def displayText(self, value, locale):
        """
        Returns the display text for the given value.

        Args:
            value: The value to display.
            locale: The locale to use for formatting.

        Returns:
            str: The display text.
        """
        if not isinstance(value, dict):
            return "Undefined"

        if sorted(value.keys()) != ["amount", "normalize", "unit"]:
            return "Faulty property"

        display = f"{value['amount']} {value['unit']}"
        return display

    def createEditor(self, parent, option, index):
        """
        Creates an editor for the given index.

        Args:
            parent: The parent widget.
            option: The style options.
            index: The index to edit.

        Returns:
            QtWidgets.QLineEdit: The editor widget.
        """
        data = index.data(QtCore.Qt.ItemDataRole.DisplayRole)
        if not data:
            return None

        editor = QtWidgets.QLineEdit(parent)
        # validator = QtGui.QDoubleValidator()
        # editor.setValidator(validator)
        return editor

    def setEditorData(self, editor: QtWidgets.QLineEdit, index: QtCore.QModelIndex):
        """
        Populates the editor with data if editing an existing field.

        Args:
            editor (QtWidgets.QLineEdit): The editor widget.
            index (QtCore.QModelIndex): The index to edit.
        """
        data = index.data(QtCore.Qt.ItemDataRole.DisplayRole)
        editor.setText(str(data["amount"]))

    def setModelData(self, editor: QtWidgets.QLineEdit, model: QtCore.QAbstractItemModel, index: QtCore.QModelIndex):
        """
        Takes the editor, reads the given value, and sets it in the model.

        Args:
            editor (QtWidgets.QLineEdit): The editor widget.
            model (QtCore.QAbstractItemModel): The model to update.
            index (QtCore.QModelIndex): The index to update.
        """
        try:
            value = float(editor.text())
            model.setData(index, value, QtCore.Qt.EditRole)
        except ValueError:
            pass


from typing import List
from logging import getLogger

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

from .method_open import MethodOpen

log = getLogger(__name__)


class MethodDuplicate(ABAction):
    """
    ABAction to duplicate a method, or node with all underlying methods to a new name specified by the user.
    """

    icon = qicons.copy
    text = "Duplicate Impact Category"

    @staticmethod
    @exception_dialogs
    def run(methods: List[tuple], level: str = None):
        # this action can handle only one selected method for now
        selected_method = methods[0]

        # check whether we're dealing with a leaf or node. If it's a node, select all underlying methods for duplication
        if level is not None and level != "leaf":
            all_methods = [
                bd.Method(method)
                for method in bd.methods
                if set(selected_method).issubset(method)
            ]
        else:
            all_methods = [bd.Method(selected_method)]

        # retrieve the new name(s) from the user and return if canceled
        dialog = TupleNameDialog.get_combined_name(
            application.main_window,
            "Impact category name",
            "Combined name:",
            selected_method,
            " - Copy",
        )
        if dialog.exec_() != TupleNameDialog.Accepted:
            return

        # for each method to be duplicated, construct a new location
        location = dialog.result_tuple
        new_names = [location + method.name[len(location) :] for method in all_methods]

        # instruct the ImpactCategoryController to duplicate the methods to the new locations
        for method, new_name in zip(all_methods, new_names):
            if new_name in methods:
                raise Exception("New method name already in use")
            method.copy(new_name)
            log.info(f"Copied method {method.name} into {new_name}")

        MethodOpen.run(new_names)


class TupleNameDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.name_label = QtWidgets.QLabel("New name")
        self.view_name = QtWidgets.QLabel()

        self.input_fields = []
        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
        )
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        layout = QtWidgets.QVBoxLayout()
        row = QtWidgets.QHBoxLayout()
        row.addWidget(self.name_label)
        row.addWidget(self.view_name)
        layout.addLayout(row)
        self.input_box = QtWidgets.QGroupBox(self)
        input_field_layout = QtWidgets.QVBoxLayout()
        self.input_box.setLayout(input_field_layout)
        layout.addWidget(self.input_box)
        layout.addWidget(self.buttons)
        self.setLayout(layout)

    @property
    def combined_names(self) -> str:
        """Reads all of the input fields in order and returns a string."""
        return ", ".join(self.result_tuple)

    @property
    def result_tuple(self) -> tuple:
        return tuple([f.text() for f in self.input_fields if f.text()])

    def changed(self) -> None:
        """
        Actions when the text within the TupleNameDialog is edited by the user
        """
        # rebuild the combined name example
        self.view_name.setText(f"'({self.combined_names})'")

        # disable the button (and its outline) when all fields are empty
        if self.combined_names == "":
            self.buttons.buttons()[0].setDefault(False)
            self.buttons.buttons()[0].setDisabled(True)
        # enable when that's not the case (anymore)
        else:
            self.buttons.buttons()[0].setDisabled(False)
            self.buttons.buttons()[0].setDefault(True)

    def add_input_field(self, text: str) -> None:
        edit = QtWidgets.QLineEdit(text, self)
        edit.textChanged.connect(self.changed)
        self.input_fields.append(edit)
        self.input_box.layout().addWidget(edit)

    @classmethod
    def get_combined_name(
        cls,
        parent: QtWidgets.QWidget,
        title: str,
        label: str,
        fields: tuple,
        extra: str = "Extra",
    ) -> "TupleNameDialog":
        """
        Set-up a TupleNameDialog pop-up with supplied title + label. Construct fields
        for each field of the supplied tuple. Last field of the tuple is appended with
        the extra string, to avoid duplicates.
        """
        obj = cls(parent)
        obj.setWindowTitle(title)
        obj.name_label.setText(label)

        # set up a field for each tuple element
        for i, field in enumerate(fields):
            field_content = str(field)

            # if it's the last element, add extra to the string
            if i + 1 == len(fields):
                field_content += extra
            obj.add_input_field(field_content)
        obj.input_box.updateGeometry()
        obj.changed()
        return obj



from qtpy.QtWidgets import QProgressDialog

from activity_browser import application
from activity_browser.mod.tqdm import qt_tqdm
from activity_browser.mod.pyprind import qt_pyprind


class ABProgressDialog(QProgressDialog):

    @classmethod
    def get_connected_dialog(cls, title: str) -> "ABProgressDialog":
        dialog = cls(application.main_window)
        dialog.setWindowTitle(title)
        dialog.setLabelText("Initializing")
        dialog.setAutoReset(False)
        dialog.setCancelButton(None)

        qt_tqdm.updated.connect(dialog._receive_update)
        qt_pyprind.updated.connect(dialog._receive_update)

        return dialog

    def _receive_update(self, title: str, value: int):
        self.setLabelText(title)
        self.setValue(value)



from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.wizards.db_export_wizard import DatabaseExportWizard


class DatabaseExport(ABAction):
    """
    ABAction to open the DatabaseExportWizard.
    """

    icon = application.style().standardIcon(QtWidgets.QStyle.SP_DriveHDIcon)
    text = "Export database..."
    tool_tip = "Export a database from this project"

    @staticmethod
    @exception_dialogs
    def run():
        DatabaseExportWizard(application.main_window).show()



# -*- coding: utf-8 -*-
from bw2data import databases
from qtpy import QtCore, QtWidgets


class DatabaseDelegate(QtWidgets.QStyledItemDelegate):
    """Nearly the same as the string delegate, but presents as
    a combobox menu containing the databases of the current project.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        editor = QtWidgets.QComboBox(parent)
        editor.insertItems(0, databases.list)
        return editor

    def setEditorData(self, editor: QtWidgets.QComboBox, index: QtCore.QModelIndex):
        """Populate the editor with data if editing an existing field."""
        value = str(index.data(QtCore.Qt.DisplayRole))
        editor.setCurrentText(value)

    def setModelData(
        self,
        editor: QtWidgets.QComboBox,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ):
        """Take the editor, read the given value and set it in the model."""
        value = editor.currentText()
        model.setData(index, value, QtCore.Qt.EditRole)
